/*70-762 Lesson 4.9 Lab 5 - Using the ALTER INDEX Statement */


Use WideWorldImportersDW;
GO

--Use the following dynamic management view (DMV) to see the OPEN rowgroups
--and determine if the columnstore index needs reorganizing especially
--after a bulk inset oata load
--NOTE:      
-- The number of multiple rowgroups depends on the degree of parallelism   
-- In SQL Server 2016, insert operations can run in parallel  

SELECT *   
FROM sys.dm_db_column_store_row_group_physical_stats   
WHERE object_id  = object_id('Fact.Order')  
ORDER BY row_group_id;

--This query calculates the FRAGMENTATION column 
--as an estimate of the efficiency of each row group
--in the current database
--To view fragmentation for a particular table,
--remove comment hyphens in WHERE clause and supply the table name

SELECT i.object_id,   
    object_name(i.object_id) AS TableName,   
    i.name AS IndexName,   
    i.index_id,   
    i.type_desc,   
    CSRowGroups.*,  
    100*(ISNULL(deleted_rows,0))/total_rows AS 'Fragmentation'  
FROM sys.indexes AS i  
JOIN sys.dm_db_column_store_row_group_physical_stats AS CSRowGroups  
    ON i.object_id = CSRowGroups.object_id AND i.index_id = CSRowGroups.index_id   
-- WHERE object_name(i.object_id) = 'table_name'   
ORDER BY object_name(i.object_id), i.name, row_group_id;

-----------------------------------------------------------------------
---------------------------------------------------------------------
-----------------------------------------------------------------------

/****AN EXAMPLE TO SHOW HOW THE REORGANIZE PROCESS WORKS*****************/
/************************************************************************/
--This example shows how ALTER INDEX REORGANIZE can force
--all deltastore rowgroups into the columnstore
--and then combine the rowgroups
--To demonstrate, a staging table containing 300,000 rows is created
--and bulk loaded into a table with clustered columnstore index
--------------------------------------------------------------------------

--1. Create the staging table containing 400,000 rows

USE master;  
GO  

DROP DATABASE IF EXISTS [ColumnstoreExample];  
GO  

--Create new database for example----------
CREATE DATABASE [ColumnstoreExample];  
GO  

USE [ColumnstoreExample];
GO

 
DROP TABLE IF EXISTS dbo.Staging;  
GO  
--Create staging table in new database
CREATE TABLE [Staging] (  
     ItemKey int NOT NULL,  
     ItemDescription nvarchar (50),  
     ItemType nvarchar(50),  
     ItemCodeAlternateKey int  
);  
GO  

--Load data into staging table 
DECLARE @loop int;  
DECLARE @ItemDescription varchar(50);  
DECLARE @ItemKey int;  
DECLARE @ItemType varchar(50);  
DECLARE @ItemCode int;  

SELECT @loop = 0;  
BEGIN TRAN  
    WHILE (@loop < 400000)   
      BEGIN  
        SELECT @ItemKey = CAST (RAND()*10000000 AS int);  
        SELECT @ItemDescription = 'Itemdesc ' + CONVERT(varchar(20), @ItemKey);  
        SELECT @ItemType = 'ItemType ' + CONVERT(varchar(20), @ItemKey);  
        SELECT @ItemCode =  CAST (RAND()*10000000 AS int);  

        INSERT INTO Staging VALUES (  
           @ItemKey,   
           @ItemDescription,   
           @ItemType,   
           @ItemCode  
        );  

        SELECT @loop = @loop + 1;  
      END  
COMMIT
GO

------------------------------------------------------
--2.  Create Target_CCI_Table with columnstore index
--    in ColumnstoreExample database


DROP TABLE IF EXISTS dbo.Target_CCI_Table;  
GO  

-- Create the table with a clustered columnstore index  
-- and the same columns as the rowstore staging table  
CREATE TABLE Target_CCI_Table (  
     ItemKey int NOT NULL,  
     ItemDescription nvarchar (50),  
     ItemType nvarchar(50),  
     ItemCodeAlternateKey int,  
     INDEX idx_Target_CCI_Table CLUSTERED COLUMNSTORE  
)  
GO

--3.  Bulk insert the staging table rows into the columnstore table
--    INSERT INTO ... SELECT performs the bulk insert
--    Insert rows in parallel employing WITH (TABLOCK) 
INSERT INTO Target_CCI_Table WITH (TABLOCK)  
SELECT TOP (400000) * FROM Staging;  
GO

--4.  Use sys.dm_db_column_store_row_group_physical_stats
--    dynamic management view (DMV) to view rowgroups

SELECT *   
FROM sys.dm_db_column_store_row_group_physical_stats   
WHERE object_id  = object_id('Target_CCI_Table')  
ORDER BY row_group_id;
GO

--5. Use ALTER INDEX REORGANIZE with the COMPRESS_ALL_ROW_GROUPS option
--   to force all rowgroups to be compressed into the columnstore.
--   All CLOSED and OPEN rowgroups are forced into the columnstore
ALTER INDEX idx_Target_CCI_Table ON Target_CCI_Table   
REORGANIZE WITH (COMPRESS_ALL_ROW_GROUPS = ON);  
GO

SELECT *   
FROM sys.dm_db_column_store_row_group_physical_stats   
WHERE object_id  = object_id('Target_CCI_Table')  
ORDER BY row_group_id;
GO


--6. Use ALTER INDEX REORGANIZE (again) to combine COMPRESSED rowgroups together. 
--   Combining small rowgroups together improves query performance.
ALTER INDEX idx_Target_CCI_Table ON Target_CCI_Table REORGANIZE;  
GO

SELECT *   
FROM sys.dm_db_column_store_row_group_physical_stats   
WHERE object_id  = object_id('Target_CCI_Table')  
ORDER BY row_group_id;
GO

---------------------------------------------------------------------------
--EXTRA: FRAGMENTATION IN B-TREE INDEXES-----------------------------------
---------------------------------------------------------------------------
--Use dynamic management views and functions
--to inspect index health
--Example: 
--Show indexes with fragmentation greater than 10 percent 
--in WideWorldImportersDW database
--For an index, one row is returned for each level of the B-tree in each partition
--NOTE: Does not return data for clustered columnstore indexes

Use WideWorldImportersDW;
GO

SELECT DB_NAME(database_id) AS 'Database',
	OBJECT_NAME(dm.object_id) AS 'Table',
	si.name AS 'Index',
	dm.index_type_desc,
	dm.avg_fragmentation_in_percent
FROM sys.dm_db_index_physical_stats
  (DB_ID(N'WideWorldImportersDW'), NULL, NULL, NULL, 'DETAILED') dm
JOIN
	sys.indexes si ON si.object_id = dm.object_id 
	AND
	si.index_id = dm.index_id
WHERE avg_fragmentation_in_percent >10
ORDER BY avg_fragmentation_in_percent DESC;
GO

-----------------------------------------------------
--Clean up code if desired
USE master;  
DROP DATABASE IF EXISTS [ColumnstoreExample];  
GO  
